@@ -43,6 +43,7 @@ enum class MessageType : uint8_t
     SetOptions,
     Exit,
     Key,
+    Paste,
 };
 
 class MsgWriter
@@ -183,6 +184,9 @@ private:
         }
     };
 
+    template<typename T>
+    struct Reader<ArrayView<T>> : Reader<Vector<std::remove_cv_t<T>, MemoryDomain::Undefined>> {};
+
     template<typename Key, typename Value, MemoryDomain domain>
     struct Reader<HashMap<Key, Value, domain>> {
         static HashMap<Key, Value, domain> read(MsgReader& reader)
@@ -256,7 +260,7 @@ public:
     }
 
     template<typename T>
-    T read()
+    auto read()
     {
         return Reader<T>::read(*this);
     }
@@ -410,6 +414,9 @@ public:
     void set_on_key(OnKeyCallback callback) override
     { m_on_key = std::move(callback); }
 
+    void set_on_paste(OnPasteCallback callback) override
+    { m_on_paste = std::move(callback); }
+
     void set_ui_options(const Options& options) override;
 
     void exit(int status);
@@ -427,6 +434,7 @@ private:
     MsgReader     m_reader;
     DisplayCoord  m_dimensions;
     OnKeyCallback m_on_key;
+    OnPasteCallback m_on_paste;
     RemoteBuffer  m_send_buffer;
 };
 
@@ -476,17 +484,25 @@ RemoteUI::RemoteUI(int socket, DisplayCoord dimensions)
                   if (not m_reader.ready())
                       continue;
 
-                   if (m_reader.type() != MessageType::Key)
+                   if (m_reader.type() == MessageType::Key)
+                   {
+                       auto key = m_reader.read<Key>();
+                       m_reader.reset();
+                       if (key.modifiers == Key::Modifiers::Resize)
+                           m_dimensions = key.coord();
+                       m_on_key(key);
+                   }
+                   else if (m_reader.type() == MessageType::Paste)
+                   {
+                       auto content = m_reader.read<String>();
+                       m_reader.reset();
+                       m_on_paste(content);
+                   }
+                   else
                    {
                        m_socket_watcher.close_fd();
                        return;
                    }
-
-                   auto key = m_reader.read<Key>();
-                   m_reader.reset();
-                   if (key.modifiers == Key::Modifiers::Resize)
-                       m_dimensions = key.coord();
-                   m_on_key(key);
               }
           }
           catch (const disconnected& err)
@@ -657,6 +673,11 @@ RemoteClient::RemoteClient(StringView session, StringView name, std::unique_ptr<
         msg.write(key);
         m_socket_watcher->events() |= FdEvents::Write;
      });
+    m_ui->set_on_paste([this](StringView content){
+        MsgWriter msg(m_send_buffer, MessageType::Paste);
+        msg.write(content);
+        m_socket_watcher->events() |= FdEvents::Write;
+     });
 
     m_socket_watcher.reset(new FDWatcher{sock, FdEvents::Read | FdEvents::Write, EventMode::Urgent,
                            [this, reader = MsgReader{}](FDWatcher& watcher, FdEvents events, EventMode) mutable {
@@ -664,6 +685,17 @@ RemoteClient::RemoteClient(StringView session, StringView name, std::unique_ptr<
         if (events & FdEvents::Write and send_data(sock, m_send_buffer))
             watcher.events() &= ~FdEvents::Write;
 
+        auto exec = [&]<typename ...Args>(void (UserInterface::*method)(Args...)) {
+            struct Impl // Use a constructor to ensure left-to-right parameter evaluation
+            {
+                Impl(UserInterface& ui, void (UserInterface::*method)(Args...), Args... args)
+                {
+                    (ui.*method)(std::forward<Args>(args)...);
+                }
+            };
+            Impl{*m_ui, method, reader.read<std::remove_cvref_t<Args>>()...};
+        };
+
         while (events & FdEvents::Read and
                not reader.ready() and fd_readable(sock))
         {
@@ -676,62 +708,34 @@ RemoteClient::RemoteClient(StringView session, StringView name, std::unique_ptr<
             switch (reader.type())
             {
             case MessageType::MenuShow:
-            {
-                auto choices = reader.read<Vector<DisplayLine>>();
-                auto anchor = reader.read<DisplayCoord>();
-                auto fg = reader.read<Face>();
-                auto bg = reader.read<Face>();
-                auto style = reader.read<MenuStyle>();
-                m_ui->menu_show(choices, anchor, fg, bg, style);
+                exec(&UserInterface::menu_show);
                 break;
-            }
             case MessageType::MenuSelect:
-                m_ui->menu_select(reader.read<int>());
+                exec(&UserInterface::menu_select);
                 break;
             case MessageType::MenuHide:
-                m_ui->menu_hide();
+                exec(&UserInterface::menu_hide);
                 break;
             case MessageType::InfoShow:
-            {
-                auto title = reader.read<DisplayLine>();
-                auto content = reader.read<DisplayLineList>();
-                auto anchor = reader.read<DisplayCoord>();
-                auto face = reader.read<Face>();
-                auto style = reader.read<InfoStyle>();
-                m_ui->info_show(title, content, anchor, face, style);
+                exec(&UserInterface::info_show);
                 break;
-            }
             case MessageType::InfoHide:
-                m_ui->info_hide();
+                exec(&UserInterface::info_hide);
                 break;
             case MessageType::Draw:
-            {
-                auto display_buffer = reader.read<DisplayBuffer>();
-                auto default_face = reader.read<Face>();
-                auto padding_face = reader.read<Face>();
-                m_ui->draw(display_buffer, default_face, padding_face);
+                exec(&UserInterface::draw);
                 break;
-            }
             case MessageType::DrawStatus:
-            {
-                auto status_line = reader.read<DisplayLine>();
-                auto mode_line = reader.read<DisplayLine>();
-                auto default_face = reader.read<Face>();
-                m_ui->draw_status(status_line, mode_line, default_face);
+                exec(&UserInterface::draw_status);
                 break;
-            }
             case MessageType::SetCursor:
-            {
-                auto mode = reader.read<CursorMode>();
-                auto coord = reader.read<DisplayCoord>();
-                m_ui->set_cursor(mode, coord);
+                exec(&UserInterface::set_cursor);
                 break;
-            }
             case MessageType::Refresh:
-                m_ui->refresh(reader.read<bool>());
+                exec(&UserInterface::refresh);
                 break;
             case MessageType::SetOptions:
-                m_ui->set_ui_options(reader.read<HashMap<String, String, MemoryDomain::Options>>());
+                exec(&UserInterface::set_ui_options);
                 break;
             case MessageType::Exit:
                 m_exit_status = reader.read<int>();
@@ -807,7 +811,7 @@ private:
                 auto* ui = new RemoteUI{sock, dimensions};
                 ClientManager::instance().create_client(
                     std::unique_ptr<UserInterface>(ui), pid, std::move(name),
-                    std::move(env_vars), init_cmds, init_coord,
+                    std::move(env_vars), init_cmds, {}, init_coord,
                     [ui](int status) { ui->exit(status); });
 
                 Server::instance().remove_accepter(this);
